package org.hibernate.cache.infinispan.impl;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.atomic.AtomicReference;

import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;

import org.hibernate.cache.CacheException;
import org.hibernate.cache.Region;
import org.hibernate.cache.RegionFactory;
import org.hibernate.cache.infinispan.util.AddressAdapter;
import org.hibernate.cache.infinispan.util.AddressAdapterImpl;
import org.hibernate.cache.infinispan.util.CacheAdapter;
import org.hibernate.cache.infinispan.util.CacheHelper;
import org.hibernate.cache.infinispan.util.FlagAdapter;
import org.infinispan.notifications.cachelistener.annotation.CacheEntryInvalidated;
import org.infinispan.notifications.cachelistener.annotation.CacheEntryModified;
import org.infinispan.notifications.cachelistener.event.CacheEntryInvalidatedEvent;
import org.infinispan.notifications.cachelistener.event.CacheEntryModifiedEvent;
import org.infinispan.notifications.cachemanagerlistener.annotation.ViewChanged;
import org.infinispan.notifications.cachemanagerlistener.event.ViewChangedEvent;
import org.infinispan.util.logging.Log;
import org.infinispan.util.logging.LogFactory;

/**
 * Support for Infinispan {@link Region}s. Handles common "utility" methods for an underlying named
 * Cache. In other words, this implementation doesn't actually read or write data. Subclasses are
 * expected to provide core cache interaction appropriate to the semantics needed.
 * 
 * @author Chris Bredesen
 * @author Galder Zamarreño
 * @since 3.5
 */
public abstract class BaseRegion implements Region {
   private enum InvalidateState {
      INVALID, CLEARING, VALID
   };
   private static final Log log = LogFactory.getLog(BaseRegion.class);
   private final String name;
   protected final CacheAdapter cacheAdapter;
   protected final AddressAdapter address;
   protected final Set<AddressAdapter> currentView = new HashSet<AddressAdapter>();
   protected final TransactionManager transactionManager;
   protected final boolean replication;
   protected final Object invalidationMutex = new Object();
   protected final AtomicReference<InvalidateState> invalidateState =
      new AtomicReference<InvalidateState>(InvalidateState.VALID);
   private final RegionFactory factory;

   public BaseRegion(CacheAdapter cacheAdapter, String name, TransactionManager transactionManager,
      RegionFactory factory)
   {
      this.cacheAdapter = cacheAdapter;
      this.name = name;
      this.transactionManager = transactionManager;
      this.replication = cacheAdapter.isClusteredReplication();
      this.address = this.cacheAdapter.getAddress();
      this.cacheAdapter.addListener(this);
      this.factory = factory;
   }

   public void start() {
      if (this.address != null) {
         synchronized (this.currentView) {
            final List<AddressAdapter> view = this.cacheAdapter.getMembers();
            if (view != null) {
               this.currentView.addAll(view);
               this.establishInternalNodes();
            }
         }
      }
   }

   /**
    * Calls to this method must be done from synchronized (currentView) blocks only!!
    */
   private void establishInternalNodes() {
      final Transaction tx = this.suspend();
      try {
         for (final AddressAdapter member : this.currentView) {
            CacheHelper.initInternalEvict(this.cacheAdapter, member);
         }
      }
      finally {
         this.resume(tx);
      }
   }

   @Override
   public String getName() {
      return this.name;
   }

   public CacheAdapter getCacheAdapter() {
      return this.cacheAdapter;
   }

   @Override
   public long getElementCountInMemory() {
      if (this.checkValid()) {
         final Set keySet = this.cacheAdapter.keySet();
         int size = this.cacheAdapter.size();
         if (CacheHelper.containsEvictAllNotification(keySet, this.address))
            size--;
         return size;
      }
      return 0;
   }

   /**
    * Not supported.
    * 
    * @return -1
    */
   @Override
   public long getElementCountOnDisk() {
      return -1;
   }

   /**
    * Not supported.
    * 
    * @return -1
    */
   @Override
   public long getSizeInMemory() {
      return -1;
   }

   @Override
   public int getTimeout() {
      return 600; // 60 seconds
   }

   @Override
   public long nextTimestamp() {
      return this.factory.nextTimestamp();
   }

   @Override
   public Map toMap() {
      if (this.checkValid()) {
         // If copying causes issues, provide a lazily loaded Map
         final Map map = new HashMap();
         final Set<Map.Entry> entries = this.cacheAdapter.toMap().entrySet();
         for (final Map.Entry entry : entries) {
            final Object key = entry.getKey();
            if (!CacheHelper.isEvictAllNotification(key)) {
               map.put(key, entry.getValue());
            }
         }
         return map;
      }
      return Collections.EMPTY_MAP;
   }

   @Override
   public void destroy() throws CacheException {
      try {
         this.cacheAdapter.clear();
      }
      finally {
         this.cacheAdapter.removeListener(this);
      }
   }

   @Override
   public boolean contains(Object key) {
      if (!this.checkValid())
         return false;
      // Reads are non-blocking in Infinispan, so not sure of the necessity of passing
      // ZERO_LOCK_ACQUISITION_TIMEOUT
      return this.cacheAdapter.withFlags(FlagAdapter.ZERO_LOCK_ACQUISITION_TIMEOUT).containsKey(key);
   }

   public AddressAdapter getAddress() {
      return this.address;
   }

   public boolean checkValid() {
      boolean valid = this.invalidateState.get() == InvalidateState.VALID;
      if (!valid) {
         synchronized (this.invalidationMutex) {
            if (this.invalidateState
               .compareAndSet(InvalidateState.INVALID, InvalidateState.CLEARING)) {
               final Transaction tx = this.suspend();
               try {
                  this.cacheAdapter.withFlags(FlagAdapter.CACHE_MODE_LOCAL,
                     FlagAdapter.ZERO_LOCK_ACQUISITION_TIMEOUT).clear();
                  this.invalidateState
                     .compareAndSet(InvalidateState.CLEARING, InvalidateState.VALID);
               }
               catch (final Exception e) {
                  if (BaseRegion.log.isTraceEnabled()) {
                     BaseRegion.log.trace("Could not invalidate region: " + e.getLocalizedMessage());
                  }
               }
               finally {
                  this.resume(tx);
               }
            }
         }
         valid = this.invalidateState.get() == InvalidateState.VALID;
      }

      return valid;
   }

   /**
    * Performs a Infinispan <code>get(Fqn, Object)</code>
    * 
    * @param key The key of the item to get
    * @param opt any option to add to the get invocation. May be <code>null</code>
    * @param suppressTimeout should any TimeoutException be suppressed?
    * @return The retrieved object
    * @throws CacheException issue managing transaction or talking to cache
    */
   protected Object get(Object key, FlagAdapter opt, boolean suppressTimeout) throws CacheException {
      if (suppressTimeout) return this.cacheAdapter.getAllowingTimeout(key);
      else return this.cacheAdapter.get(key);
   }

   public Object getOwnerForPut() {
      Transaction tx = null;
      try {
         if (this.transactionManager != null) {
            tx = this.transactionManager.getTransaction();
         }
      }
      catch (final SystemException se) {
         throw new CacheException("Could not obtain transaction", se);
      }
      return tx == null ? Thread.currentThread() : tx;
   }

   /**
    * Tell the TransactionManager to suspend any ongoing transaction.
    * 
    * @return the transaction that was suspended, or <code>null</code> if there wasn't one
    */
   public Transaction suspend() {
   	// we do not want to suspend our on going transactions
   	// as the underlying cache is transactional.
   	// Suspending it would cause the system to auto create transaction,
   	// commit and replicate/invalidate the data.
   	// We only want that to happen with our main transaction.
      return null;
   }

   /**
    * Tell the TransactionManager to resume the given transaction
    * 
    * @param tx the transaction to suspend. May be <code>null</code>.
    */
   public void resume(Transaction tx) {
	   // see suspend.
      assert tx == null;
      // assert null because it is expected to be
      // transaction that got suspended, and since we do not suspend them...
   }

   @CacheEntryModified
   public void entryModified(CacheEntryModifiedEvent event) {
      this.handleEvictAllModification(event);
   }

   protected boolean handleEvictAllModification(CacheEntryModifiedEvent event) {
      if (!event.isPre() && (this.replication || event.isOriginLocal()) &&
         CacheHelper.isEvictAllNotification(event.getKey(), event.getValue())) {
         // bad method signature in the trace that we removed.
         this.invalidateState.set(InvalidateState.INVALID);
         return true;
      }
      return false;
   }

   @CacheEntryInvalidated
   public void entryInvalidated(CacheEntryInvalidatedEvent event) {
      // bad method signature in the trace that we removed.
      this.handleEvictAllInvalidation(event);
   }

   protected boolean handleEvictAllInvalidation(CacheEntryInvalidatedEvent event) {
      if (!event.isPre() && CacheHelper.isEvictAllNotification(event.getKey())) {
         // bad method signature in the trace that we removed.
         this.invalidateState.set(InvalidateState.INVALID);
         return true;
      }
      return false;
   }

   @ViewChanged
   public void viewChanged(ViewChangedEvent event) {
      synchronized (this.currentView) {
         final List<AddressAdapter> view =
            AddressAdapterImpl.toAddressAdapter(event.getNewMembers());
         if (view != null) {
            this.currentView.addAll(view);
            this.establishInternalNodes();
         }
      }
   }

}