/*
 * Hibernate, Relational Persistence for Idiomatic Java Copyright (c) 2010, Red Hat Inc. or
 * third-party contributors as indicated by the @author tags or express copyright attribution
 * statements applied by the authors. All third-party contributions are distributed under license by
 * Red Hat Inc. This copyrighted material is made available to anyone wishing to use, modify, copy,
 * or redistribute it subject to the terms and conditions of the GNU Lesser General Public License,
 * as published by the Free Software Foundation. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details. You
 * should have received a copy of the GNU Lesser General Public License along with this
 * distribution; if not, write to: Free Software Foundation, Inc. 51 Franklin Street, Fifth Floor
 * Boston, MA 02110-1301 USA
 */
package org.hibernate.engine;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.hibernate.AssertionFailure;
import org.hibernate.HibernateException;
import org.hibernate.action.AfterTransactionCompletionProcess;
import org.hibernate.action.BeforeTransactionCompletionProcess;
import org.hibernate.action.BulkOperationCleanupAction;
import org.hibernate.action.CollectionRecreateAction;
import org.hibernate.action.CollectionRemoveAction;
import org.hibernate.action.CollectionUpdateAction;
import org.hibernate.action.EntityDeleteAction;
import org.hibernate.action.EntityIdentityInsertAction;
import org.hibernate.action.EntityInsertAction;
import org.hibernate.action.EntityUpdateAction;
import org.hibernate.action.Executable;
import org.hibernate.cache.CacheException;
import org.hibernate.type.Type;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Responsible for maintaining the queue of actions related to events. </p> The ActionQueue holds
 * the DML operations queued as part of a session's transactional-write-behind semantics. DML
 * operations are queued here until a flush forces them to be executed against the database.
 * 
 * @author Steve Ebersole
 */
public class ActionQueue {

   private static final Logger log = LoggerFactory.getLogger(ActionQueue.class);
   private static final int INIT_QUEUE_LIST_SIZE = 5;

   private final SessionImplementor session;

   // Object insertions, updates, and deletions have list semantics because
   // they must happen in the right order so as to respect referential
   // integrity
   private ArrayList insertions;
   private ArrayList deletions;
   private ArrayList updates;
   // Actually the semantics of the next three are really "Bag"
   // Note that, unlike objects, collection insertions, updates,
   // deletions are not really remembered between flushes. We
   // just re-use the same Lists for convenience.
   private ArrayList collectionCreations;
   private ArrayList collectionUpdates;
   private ArrayList collectionRemovals;

   private AfterTransactionCompletionProcessQueue afterTransactionProcesses;
   private BeforeTransactionCompletionProcessQueue beforeTransactionProcesses;

   /**
    * Constructs an action queue bound to the given session.
    * 
    * @param session The session "owning" this queue.
    */
   public ActionQueue(SessionImplementor session) {
      this.session = session;
      this.init();
   }

   private void init() {
      this.insertions = new ArrayList(ActionQueue.INIT_QUEUE_LIST_SIZE);
      this.deletions = new ArrayList(ActionQueue.INIT_QUEUE_LIST_SIZE);
      this.updates = new ArrayList(ActionQueue.INIT_QUEUE_LIST_SIZE);

      this.collectionCreations = new ArrayList(ActionQueue.INIT_QUEUE_LIST_SIZE);
      this.collectionRemovals = new ArrayList(ActionQueue.INIT_QUEUE_LIST_SIZE);
      this.collectionUpdates = new ArrayList(ActionQueue.INIT_QUEUE_LIST_SIZE);

      this.afterTransactionProcesses = new AfterTransactionCompletionProcessQueue(this.session);
      this.beforeTransactionProcesses = new BeforeTransactionCompletionProcessQueue(this.session);
   }

   public void clear() {
      this.updates.clear();
      this.insertions.clear();
      this.deletions.clear();

      this.collectionCreations.clear();
      this.collectionRemovals.clear();
      this.collectionUpdates.clear();
   }

   public void addAction(EntityInsertAction action) {
      this.insertions.add(action);
   }

   public void addAction(EntityDeleteAction action) {
      this.deletions.add(action);
   }

   public void addAction(EntityUpdateAction action) {
      this.updates.add(action);
   }

   public void addAction(CollectionRecreateAction action) {
      this.collectionCreations.add(action);
   }

   public void addAction(CollectionRemoveAction action) {
      this.collectionRemovals.add(action);
   }

   public void addAction(CollectionUpdateAction action) {
      this.collectionUpdates.add(action);
   }

   public void addAction(EntityIdentityInsertAction insert) {
      this.insertions.add(insert);
   }

   public void addAction(BulkOperationCleanupAction cleanupAction) {
      this.registerCleanupActions(cleanupAction);
   }

   public void registerProcess(AfterTransactionCompletionProcess process) {
      this.afterTransactionProcesses.register(process);
   }

   public void registerProcess(BeforeTransactionCompletionProcess process) {
      this.beforeTransactionProcesses.register(process);
   }

   /**
    * Perform all currently queued entity-insertion actions.
    * 
    * @throws HibernateException error executing queued insertion actions.
    */
   public void executeInserts() throws HibernateException {
      this.executeActions(this.insertions);
   }

   /**
    * Perform all currently queued actions.
    * 
    * @throws HibernateException error executing queued actions.
    */
   public void executeActions() throws HibernateException {
      this.executeActions(this.insertions);
      this.executeActions(this.updates);
      this.executeActions(this.collectionRemovals);
      this.executeActions(this.collectionUpdates);
      this.executeActions(this.collectionCreations);
      this.executeActions(this.deletions);
   }

   /**
    * Prepares the internal action queues for execution.
    * 
    * @throws HibernateException error preparing actions.
    */
   public void prepareActions() throws HibernateException {
      this.prepareActions(this.collectionRemovals);
      this.prepareActions(this.collectionUpdates);
      this.prepareActions(this.collectionCreations);
   }

   /**
    * Performs cleanup of any held cache softlocks.
    * 
    * @param success Was the transaction successful.
    */
   public void afterTransactionCompletion(boolean success) {
      this.afterTransactionProcesses.afterTransactionCompletion(success);
   }

   /**
    * Execute any registered {@link BeforeTransactionCompletionProcess}
    */
   public void beforeTransactionCompletion() {
      this.beforeTransactionProcesses.beforeTransactionCompletion();
   }

   /**
    * Check whether the given tables/query-spaces are to be executed against given the currently
    * queued actions.
    * 
    * @param tables The table/query-spaces to check.
    * 
    * @return True if we contain pending actions against any of the given tables; false otherwise.
    */
   public boolean areTablesToBeUpdated(Set tables) {
      return ActionQueue.areTablesToUpdated(this.updates, tables) ||
         ActionQueue.areTablesToUpdated(this.insertions, tables) ||
         ActionQueue.areTablesToUpdated(this.deletions, tables) ||
         ActionQueue.areTablesToUpdated(this.collectionUpdates, tables) ||
         ActionQueue.areTablesToUpdated(this.collectionCreations, tables) ||
         ActionQueue.areTablesToUpdated(this.collectionRemovals, tables);
   }

   /**
    * Check whether any insertion or deletion actions are currently queued.
    * 
    * @return True if insertions or deletions are currently queued; false otherwise.
    */
   public boolean areInsertionsOrDeletionsQueued() {
      return (this.insertions.size() > 0 || this.deletions.size() > 0);
   }

   private static boolean areTablesToUpdated(List executables, Set tablespaces) {
      final int size = executables.size();
      for (int j = 0; j < size; j++) {
         final Serializable[] spaces = ((Executable)executables.get(j)).getPropertySpaces();
         for (int i = 0; i < spaces.length; i++) {
            if (tablespaces.contains(spaces[i])) {
               if (ActionQueue.log.isDebugEnabled()) {
                  ActionQueue.log.debug("changes must be flushed to space: " + spaces[i]);
               }
               return true;
            }
         }
      }
      return false;
   }

   private void executeActions(List list) throws HibernateException {
      final int size = list.size();
      for (int i = 0; i < size; i++) {
         this.execute((Executable)list.get(i));
      }
      list.clear();
      this.session.getBatcher().executeBatch();
   }

   public void execute(Executable executable) {
      try {
         executable.execute();
      }
      finally {
         this.registerCleanupActions(executable);
      }
   }

   private void registerCleanupActions(Executable executable) {
      this.beforeTransactionProcesses.register(executable.getBeforeTransactionCompletionProcess());
      if (this.session.getFactory().getSettings().isQueryCacheEnabled()) {
         final String[] spaces = (String[])executable.getPropertySpaces();
         // see method involved for documentation.
         this.beforeTransactionProcesses.addSpacesToInvalidate(spaces);
         this.session.getFactory().getUpdateTimestampsCache().preinvalidate(spaces);
      }
      this.afterTransactionProcesses.register(executable.getAfterTransactionCompletionProcess());
   }

   private void prepareActions(List queue) throws HibernateException {
      final int size = queue.size();
      for (int i = 0; i < size; i++) {
         final Executable executable = (Executable)queue.get(i);
         executable.beforeExecutions();
      }
   }

   /**
    * Returns a string representation of the object.
    * 
    * @return a string representation of the object.
    */
   @Override
   public String toString() {
      return new StringBuffer()
         .append("ActionQueue[insertions=").append(this.insertions)
         .append(" updates=").append(this.updates)
         .append(" deletions=").append(this.deletions)
         .append(" collectionCreations=").append(this.collectionCreations)
         .append(" collectionRemovals=").append(this.collectionRemovals)
         .append(" collectionUpdates=").append(this.collectionUpdates)
         .append("]")
         .toString();
   }

   public int numberOfCollectionRemovals() {
      return this.collectionRemovals.size();
   }

   public int numberOfCollectionUpdates() {
      return this.collectionUpdates.size();
   }

   public int numberOfCollectionCreations() {
      return this.collectionCreations.size();
   }

   public int numberOfDeletions() {
      return this.deletions.size();
   }

   public int numberOfUpdates() {
      return this.updates.size();
   }

   public int numberOfInsertions() {
      return this.insertions.size();
   }

   public void sortCollectionActions() {
      if (this.session.getFactory().getSettings().isOrderUpdatesEnabled()) {
         // sort the updates by fk
         java.util.Collections.sort(this.collectionCreations);
         java.util.Collections.sort(this.collectionUpdates);
         java.util.Collections.sort(this.collectionRemovals);
      }
   }

   public void sortActions() {
      if (this.session.getFactory().getSettings().isOrderUpdatesEnabled()) {
         // sort the updates by pk
         java.util.Collections.sort(this.updates);
      }
      if (this.session.getFactory().getSettings().isOrderInsertsEnabled()) {
         this.sortInsertActions();
      }
   }

   /**
    * Order the {@link #insertions} queue such that we group inserts against the same entity
    * together (without violating constraints). The original order is generated by cascade order,
    * which in turn is based on the directionality of foreign-keys. So even though we will be
    * changing the ordering here, we need to make absolutely certain that we do not circumvent this
    * FK ordering to the extent of causing constraint violations
    */
   private void sortInsertActions() {
      new InsertActionSorter().sort();
   }

   public ArrayList cloneDeletions() {
      return (ArrayList)this.deletions.clone();
   }

   public void clearFromFlushNeededCheck(int previousCollectionRemovalSize) {
      this.collectionCreations.clear();
      this.collectionUpdates.clear();
      this.updates.clear();
      // collection deletions are a special case since update() can add
      // deletions of collections not loaded by the session.
      for (int i = this.collectionRemovals.size() - 1; i >= previousCollectionRemovalSize; i--) {
         this.collectionRemovals.remove(i);
      }
   }

   public boolean hasAfterTransactionActions() {
      return this.afterTransactionProcesses.processes.size() > 0;
   }

   public boolean hasBeforeTransactionActions() {
      return this.beforeTransactionProcesses.processes.size() > 0;
   }

   public boolean hasAnyQueuedActions() {
      return this.updates.size() > 0 ||
         this.insertions.size() > 0 ||
         this.deletions.size() > 0 ||
         this.collectionUpdates.size() > 0 ||
         this.collectionRemovals.size() > 0 ||
         this.collectionCreations.size() > 0;
   }

   /**
    * Used by the owning session to explicitly control serialization of the action queue
    * 
    * @param oos The stream to which the action queue should get written
    * 
    * @throws IOException Indicates an error writing to the stream
    */
   public void serialize(ObjectOutputStream oos) throws IOException {
      ActionQueue.log.trace("serializing action-queue");

      int queueSize = this.insertions.size();
      ActionQueue.log.trace("starting serialization of [" + queueSize + "] insertions entries");
      oos.writeInt(queueSize);
      for (int i = 0; i < queueSize; i++) {
         oos.writeObject(this.insertions.get(i));
      }

      queueSize = this.deletions.size();
      ActionQueue.log.trace("starting serialization of [" + queueSize + "] deletions entries");
      oos.writeInt(queueSize);
      for (int i = 0; i < queueSize; i++) {
         oos.writeObject(this.deletions.get(i));
      }

      queueSize = this.updates.size();
      ActionQueue.log.trace("starting serialization of [" + queueSize + "] updates entries");
      oos.writeInt(queueSize);
      for (int i = 0; i < queueSize; i++) {
         oos.writeObject(this.updates.get(i));
      }

      queueSize = this.collectionUpdates.size();
      ActionQueue.log.trace("starting serialization of [" + queueSize +
         "] collectionUpdates entries");
      oos.writeInt(queueSize);
      for (int i = 0; i < queueSize; i++) {
         oos.writeObject(this.collectionUpdates.get(i));
      }

      queueSize = this.collectionRemovals.size();
      ActionQueue.log.trace("starting serialization of [" + queueSize +
         "] collectionRemovals entries");
      oos.writeInt(queueSize);
      for (int i = 0; i < queueSize; i++) {
         oos.writeObject(this.collectionRemovals.get(i));
      }

      queueSize = this.collectionCreations.size();
      ActionQueue.log.trace("starting serialization of [" + queueSize +
         "] collectionCreations entries");
      oos.writeInt(queueSize);
      for (int i = 0; i < queueSize; i++) {
         oos.writeObject(this.collectionCreations.get(i));
      }
   }

   /**
    * Used by the owning session to explicitly control deserialization of the action queue
    * 
    * @param ois The stream from which to read the action queue
    * @param session The session to which the action queue belongs
    * 
    * @return The deserialized action queue
    * 
    * @throws IOException indicates a problem reading from the stream
    * @throws ClassNotFoundException Generally means we were unable to locate user classes.
    */
   public static ActionQueue deserialize(
      ObjectInputStream ois,
      SessionImplementor session) throws IOException, ClassNotFoundException
   {
      ActionQueue.log.trace("deserializing action-queue");
      final ActionQueue rtn = new ActionQueue(session);

      int queueSize = ois.readInt();
      ActionQueue.log.trace("starting deserialization of [" + queueSize + "] insertions entries");
      rtn.insertions = new ArrayList(queueSize);
      for (int i = 0; i < queueSize; i++) {
         rtn.insertions.add(ois.readObject());
      }

      queueSize = ois.readInt();
      ActionQueue.log.trace("starting deserialization of [" + queueSize + "] deletions entries");
      rtn.deletions = new ArrayList(queueSize);
      for (int i = 0; i < queueSize; i++) {
         rtn.deletions.add(ois.readObject());
      }

      queueSize = ois.readInt();
      ActionQueue.log.trace("starting deserialization of [" + queueSize + "] updates entries");
      rtn.updates = new ArrayList(queueSize);
      for (int i = 0; i < queueSize; i++) {
         rtn.updates.add(ois.readObject());
      }

      queueSize = ois.readInt();
      ActionQueue.log.trace("starting deserialization of [" + queueSize +
         "] collectionUpdates entries");
      rtn.collectionUpdates = new ArrayList(queueSize);
      for (int i = 0; i < queueSize; i++) {
         rtn.collectionUpdates.add(ois.readObject());
      }

      queueSize = ois.readInt();
      ActionQueue.log.trace("starting deserialization of [" + queueSize +
         "] collectionRemovals entries");
      rtn.collectionRemovals = new ArrayList(queueSize);
      for (int i = 0; i < queueSize; i++) {
         rtn.collectionRemovals.add(ois.readObject());
      }

      queueSize = ois.readInt();
      ActionQueue.log.trace("starting deserialization of [" + queueSize +
         "] collectionCreations entries");
      rtn.collectionCreations = new ArrayList(queueSize);
      for (int i = 0; i < queueSize; i++) {
         rtn.collectionCreations.add(ois.readObject());
      }
      return rtn;
   }

   private static class BeforeTransactionCompletionProcessQueue {
      private final SessionImplementor session;
      private final List processes = new ArrayList();
      // moved that set from AfterTransactionCompletionProcessQueue, see beforeTransactionCompletion()
      private final Set querySpacesToInvalidate = new HashSet();

      private BeforeTransactionCompletionProcessQueue(SessionImplementor session) {
         this.session = session;
      }

      public void register(BeforeTransactionCompletionProcess process) {
         if (process == null) {
            return;
         }
         this.processes.add(process);
      }

      public void addSpaceToInvalidate(String space) {
			// moved from AfterTransactionCompletionProcessQueue, see beforeTransactionCompletion(). 
         this.querySpacesToInvalidate.add(space);
      }

      public void addSpacesToInvalidate(String[] spaces) {
			// moved from AfterTransactionCompletionProcessQueue, see beforeTransactionCompletion().
         if (spaces == null) {
            return;
         }
         for (int i = 0, max = spaces.length; i < max; i++) {
            this.addSpaceToInvalidate(spaces[i]);
         }
      }

      public void beforeTransactionCompletion() {
         final int size = this.processes.size();
         for (int i = 0; i < size; i++) {
            try {
               final BeforeTransactionCompletionProcess process =
                  (BeforeTransactionCompletionProcess)this.processes.get(i);
               process.doBeforeTransactionCompletion(this.session);
            }
            catch (final HibernateException he) {
               throw he;
            }
            catch (final Exception e) {
               throw new AssertionFailure("Unable to perform beforeTransactionCompletion callback",
                  e);
            }
         }
         this.processes.clear();

			// this used to be performed in AfterTransactionCompletionProcessQueue.
			// However, it used to cause issues as it results in "putting" elements into the transactional
			// cache after the transaction has been committed...
			// It has been moved to be "put" prior to having fully committed, as it should produce the same result
			// in case of success and in case of failures, the cache being transactional it should rollback.
         if (this.session.getFactory().getSettings().isQueryCacheEnabled()) {
            this.session
               .getFactory()
               .getUpdateTimestampsCache()
               .invalidate(
                  (String[])this.querySpacesToInvalidate
                     .toArray(new String[this.querySpacesToInvalidate.size()])
               );
         }
         this.querySpacesToInvalidate.clear();
      }
   }

   private static class AfterTransactionCompletionProcessQueue {
      private final SessionImplementor session;
      private final List processes = new ArrayList(ActionQueue.INIT_QUEUE_LIST_SIZE * 3);

      private AfterTransactionCompletionProcessQueue(SessionImplementor session) {
         this.session = session;
      }

      public void register(AfterTransactionCompletionProcess process) {
         if (process == null) {
            return;
         }
         this.processes.add(process);
      }

      public void afterTransactionCompletion(boolean success) {
         final int size = this.processes.size();
         for (int i = 0; i < size; i++) {
            try {
               final AfterTransactionCompletionProcess process =
                  (AfterTransactionCompletionProcess)this.processes.get(i);
               process.doAfterTransactionCompletion(success, this.session);
            }
            catch (final CacheException ce) {
               ActionQueue.log.error("could not release a cache lock", ce);
               // continue loop
            }
            catch (final Exception e) {
               throw new AssertionFailure("Exception releasing cache locks", e);
            }
         }
         this.processes.clear();
      }
   }

   /**
    * Sorts the insert actions using more hashes.
    * 
    * @author Jay Erb
    */
   private class InsertActionSorter {
      // the mapping of entity names to their latest batch numbers.
      private final HashMap latestBatches = new HashMap();
      private final HashMap entityBatchNumber;

      // the map of batch numbers to EntityInsertAction lists
      private final HashMap actionBatches = new HashMap();

      public InsertActionSorter() {
         // optimize the hash size to eliminate a rehash.
         this.entityBatchNumber = new HashMap(ActionQueue.this.insertions.size() + 1, 1.0f);
      }

      /**
       * Sort the insert actions.
       */
      public void sort() {

         // the list of entity names that indicate the batch number
         for (final Iterator actionItr = ActionQueue.this.insertions.iterator(); actionItr.hasNext();) {
            final EntityInsertAction action = (EntityInsertAction)actionItr.next();
            // remove the current element from insertions. It will be added back later.
            final String entityName = action.getEntityName();

            // the entity associated with the current action.
            final Object currentEntity = action.getInstance();

            Integer batchNumber;
            if (this.latestBatches.containsKey(entityName)) {
               // There is already an existing batch for this type of entity.
               // Check to see if the latest batch is acceptable.
               batchNumber = this.findBatchNumber(action, entityName);
            }
            else {
               // add an entry for this type of entity.
               // we can be assured that all referenced entities have already
               // been processed,
               // so specify that this entity is with the latest batch.
               // doing the batch number before adding the name to the list is
               // a faster way to get an accurate number.

               batchNumber = new Integer(this.actionBatches.size());
               this.latestBatches.put(entityName, batchNumber);
            }
            this.entityBatchNumber.put(currentEntity, batchNumber);
            this.addToBatch(batchNumber, action);
         }
         ActionQueue.this.insertions.clear();

         // now rebuild the insertions list. There is a batch for each entry in the name list.
         for (int i = 0; i < this.actionBatches.size(); i++) {
            final List batch = (List)this.actionBatches.get(new Integer(i));
            for (final Iterator batchItr = batch.iterator(); batchItr.hasNext();) {
               final EntityInsertAction action = (EntityInsertAction)batchItr.next();
               ActionQueue.this.insertions.add(action);
            }
         }
      }

      /**
       * Finds an acceptable batch for this entity to be a member as part of the
       * {@link InsertActionSorter}
       * 
       * @param action The action being sorted
       * @param entityName The name of the entity affected by the action
       * 
       * @return An appropriate batch number; todo document this process better
       */
      private Integer findBatchNumber(
         EntityInsertAction action,
         String entityName)
      {
         // loop through all the associated entities and make sure they have been
         // processed before the latest
         // batch associated with this entity type.

         // the current batch number is the latest batch for this entity type.
         Integer latestBatchNumberForType = (Integer)this.latestBatches.get(entityName);

         // loop through all the associations of the current entity and make sure that they are
         // processed
         // before the current batch number
         final Object[] propertyValues = action.getState();
         final Type[] propertyTypes = action.getPersister().getClassMetadata()
            .getPropertyTypes();

         for (int i = 0; i < propertyValues.length; i++) {
            final Object value = propertyValues[i];
            final Type type = propertyTypes[i];
            if (type.isEntityType() && value != null) {
               // find the batch number associated with the current association, if any.
               final Integer associationBatchNumber = (Integer)this.entityBatchNumber.get(value);
               if (associationBatchNumber != null &&
                  associationBatchNumber.compareTo(latestBatchNumberForType) > 0) {
                  // create a new batch for this type. The batch number is the number of current
                  // batches.
                  latestBatchNumberForType = new Integer(this.actionBatches.size());
                  this.latestBatches.put(entityName, latestBatchNumberForType);
                  // since this entity will now be processed in the latest possible batch,
                  // we can be assured that it will come after all other associations,
                  // there's not need to continue checking.
                  break;
               }
            }
         }
         return latestBatchNumberForType;
      }

      private void addToBatch(Integer batchNumber, EntityInsertAction action) {
         List actions = (List)this.actionBatches.get(batchNumber);

         if (actions == null) {
            actions = new LinkedList();
            this.actionBatches.put(batchNumber, actions);
         }
         actions.add(action);
      }

   }

}
