/*
 * JBoss, Home of Professional Open Source Copyright 2009 Red Hat Inc. and/or its affiliates and
 * other contributors as indicated by the @author tags. All rights reserved. See the copyright.txt
 * in the distribution for a full listing of individual contributors. This is free software; you can
 * redistribute it and/or modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This software is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU Lesser General Public License for more details. You should have received a
 * copy of the GNU Lesser General Public License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA, or see the FSF
 * site: http://www.fsf.org.
 */
package org.infinispan;

import static java.util.concurrent.TimeUnit.MILLISECONDS;
import static org.infinispan.context.Flag.FAIL_SILENTLY;
import static org.infinispan.context.Flag.FORCE_ASYNCHRONOUS;
import static org.infinispan.context.Flag.PUT_FOR_EXTERNAL_READ;
import static org.infinispan.context.Flag.ZERO_LOCK_ACQUISITION_TIMEOUT;

import java.util.Arrays;
import java.util.Collection;
import java.util.EnumSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.TimeUnit;

import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import javax.transaction.xa.XAResource;

import org.infinispan.atomic.Delta;
import org.infinispan.batch.BatchContainer;
import org.infinispan.commands.CommandsFactory;
import org.infinispan.commands.VisitableCommand;
import org.infinispan.commands.control.LockControlCommand;
import org.infinispan.commands.read.EntrySetCommand;
import org.infinispan.commands.read.GetKeyValueCommand;
import org.infinispan.commands.read.KeySetCommand;
import org.infinispan.commands.read.SizeCommand;
import org.infinispan.commands.read.ValuesCommand;
import org.infinispan.commands.write.ApplyDeltaCommand;
import org.infinispan.commands.write.ClearCommand;
import org.infinispan.commands.write.EvictCommand;
import org.infinispan.commands.write.PutKeyValueCommand;
import org.infinispan.commands.write.PutMapCommand;
import org.infinispan.commands.write.RemoveCommand;
import org.infinispan.commands.write.ReplaceCommand;
import org.infinispan.config.Configuration;
import org.infinispan.config.Configuration.CacheMode;
import org.infinispan.config.ConfigurationException;
import org.infinispan.container.DataContainer;
import org.infinispan.container.entries.InternalCacheEntry;
import org.infinispan.context.Flag;
import org.infinispan.context.InvocationContext;
import org.infinispan.context.InvocationContextContainer;
import org.infinispan.context.impl.TxInvocationContext;
import org.infinispan.distribution.DistributionManager;
import org.infinispan.eviction.EvictionManager;
import org.infinispan.factories.ComponentRegistry;
import org.infinispan.factories.KnownComponentNames;
import org.infinispan.factories.annotations.ComponentName;
import org.infinispan.factories.annotations.Inject;
import org.infinispan.factories.annotations.SurvivesRestarts;
import org.infinispan.interceptors.InterceptorChain;
import org.infinispan.interceptors.base.CommandInterceptor;
import org.infinispan.jmx.annotations.MBean;
import org.infinispan.jmx.annotations.ManagedAttribute;
import org.infinispan.jmx.annotations.ManagedOperation;
import org.infinispan.lifecycle.ComponentStatus;
import org.infinispan.manager.BasicCacheContainer;
import org.infinispan.manager.EmbeddedCacheManager;
import org.infinispan.marshall.MarshalledValue;
import org.infinispan.marshall.StreamingMarshaller;
import org.infinispan.notifications.cachelistener.CacheNotifier;
import org.infinispan.remoting.responses.ResponseGenerator;
import org.infinispan.remoting.rpc.RpcManager;
import org.infinispan.statetransfer.StateTransferManager;
import org.infinispan.stats.Stats;
import org.infinispan.stats.StatsImpl;
import org.infinispan.transaction.TransactionCoordinator;
import org.infinispan.transaction.TransactionTable;
import org.infinispan.transaction.xa.TransactionXaAdapter;
import org.infinispan.transaction.xa.recovery.RecoveryManager;
import org.infinispan.util.Util;
import org.infinispan.util.concurrent.AbstractInProcessNotifyingFuture;
import org.infinispan.util.concurrent.DeferredReturnFuture;
import org.infinispan.util.concurrent.NotifyingFuture;
import org.infinispan.util.concurrent.NotifyingNotifiableFuture;
import org.infinispan.util.concurrent.locks.LockManager;
import org.infinispan.util.logging.Log;
import org.infinispan.util.logging.LogFactory;
import org.rhq.helpers.pluginAnnotations.agent.DataType;
import org.rhq.helpers.pluginAnnotations.agent.DisplayType;
import org.rhq.helpers.pluginAnnotations.agent.Metric;
import org.rhq.helpers.pluginAnnotations.agent.Operation;

/**
 * @author Mircea.Markus@jboss.com
 * @author Galder Zamarreño
 * @author Sanne Grinovero
 * @author <a href="http://gleamynode.net/">Trustin Lee</a>
 * @since 4.0
 */
@SurvivesRestarts
@MBean(objectName = CacheImpl.OBJECT_NAME,
   description = "Component that represents an individual cache instance.")
public class CacheImpl<K, V> extends CacheSupport<K, V> implements AdvancedCache<K, V> {
   public static final String OBJECT_NAME = "Cache";
   protected InvocationContextContainer icc;
   protected CommandsFactory commandsFactory;
   protected InterceptorChain invoker;
   protected Configuration config;
   protected CacheNotifier notifier;
   protected BatchContainer batchContainer;
   protected ComponentRegistry componentRegistry;
   protected TransactionManager transactionManager;
   protected RpcManager rpcManager;
   protected StreamingMarshaller marshaller;
   private final String name;
   private EvictionManager evictionManager;
   private DataContainer dataContainer;
   private static final Log log = LogFactory.getLog(CacheImpl.class);
   private static final boolean trace = CacheImpl.log.isTraceEnabled();
   private EmbeddedCacheManager cacheManager;
   // this is never used here but should be injected - this is a hack to make sure the
   // StateTransferManager is properly constructed if needed.
   private StateTransferManager stateTransferManager;
   // as above for ResponseGenerator
   private ResponseGenerator responseGenerator;
   private LockManager lockManager;
   private DistributionManager distributionManager;
   private ExecutorService asyncExecutor;
   private TransactionTable txTable;
   private RecoveryManager recoveryManager;
   private TransactionCoordinator txCoordinator;

   public CacheImpl(String name) {
      this.name = name;
   }

   @Inject
   public void injectDependencies(EvictionManager evictionManager,
      InvocationContextContainer icc,
      CommandsFactory commandsFactory,
      InterceptorChain interceptorChain,
      Configuration configuration,
      CacheNotifier notifier,
      ComponentRegistry componentRegistry,
      TransactionManager transactionManager,
      BatchContainer batchContainer,
      RpcManager rpcManager, DataContainer dataContainer,
      @ComponentName(KnownComponentNames.CACHE_MARSHALLER) StreamingMarshaller marshaller,
      ResponseGenerator responseGenerator,
      DistributionManager distributionManager,
      EmbeddedCacheManager cacheManager, StateTransferManager stateTransferManager,
      @ComponentName(KnownComponentNames.ASYNC_TRANSPORT_EXECUTOR) ExecutorService asyncExecutor,
      TransactionTable txTable, RecoveryManager recoveryManager,
      TransactionCoordinator txCoordinator,
      LockManager lockManager)
   {
      this.commandsFactory = commandsFactory;
      this.invoker = interceptorChain;
      this.config = configuration;
      this.notifier = notifier;
      this.componentRegistry = componentRegistry;
      this.transactionManager = transactionManager;
      this.batchContainer = batchContainer;
      this.rpcManager = rpcManager;
      this.evictionManager = evictionManager;
      this.dataContainer = dataContainer;
      this.marshaller = marshaller;
      this.cacheManager = cacheManager;
      this.responseGenerator = responseGenerator;
      this.stateTransferManager = stateTransferManager;
      this.icc = icc;
      this.distributionManager = distributionManager;
      this.asyncExecutor = asyncExecutor;
      this.txTable = txTable;
      this.recoveryManager = recoveryManager;
      this.txCoordinator = txCoordinator;
      this.lockManager = lockManager;
   }

   private void assertKeyNotNull(Object key) {
      if (key == null) {
         throw new NullPointerException("Null keys are not supported!");
      }
   }

   private void assertKeysNotNull(Map<?, ?> data) {
      if (data == null) {
         throw new NullPointerException("Expected map cannot be null");
      }
      for (final Object key : data.keySet()) {
         if (key == null) {
            throw new NullPointerException("Null keys are not supported!");
         }
      }
   }

   @Override
   public final boolean remove(Object key, Object value) {
      return this.remove(key, value, null, null);
   }

   final boolean remove(Object key, Object value, EnumSet<Flag> explicitFlags,
      ClassLoader explicitClassLoader)
   {
      this.assertKeyNotNull(key);
      final InvocationContext ctx =
         this.getInvocationContextWithImplicitTransaction(explicitFlags, explicitClassLoader);
      final RemoveCommand command =
         this.commandsFactory.buildRemoveCommand(key, value, ctx.getFlags());
      return (Boolean)this.executeCommandAndCommitIfNeeded(ctx, command);
   }

   @Override
   public final int size() {
      return this.size(null, null);
   }

   final int size(EnumSet<Flag> explicitFlags, ClassLoader explicitClassLoader) {
      final SizeCommand command = this.commandsFactory.buildSizeCommand();
      return (Integer)this.invoker.invoke(
         this.getInvocationContextForRead(null, explicitFlags, explicitClassLoader), command);
   }

   @Override
   public final boolean isEmpty() {
      return this.size() == 0;
   }

   final boolean isEmpty(EnumSet<Flag> explicitFlags, ClassLoader explicitClassLoader) {
      return this.size(explicitFlags, explicitClassLoader) == 0;
   }

   @Override
   public final boolean containsKey(Object key) {
      return this.containsKey(key, null, null);
   }

   final boolean
      containsKey(Object key, EnumSet<Flag> explicitFlags, ClassLoader explicitClassLoader)
   {
      this.assertKeyNotNull(key);
      final InvocationContext ctx =
         this.getInvocationContextForRead(null, explicitFlags, explicitClassLoader);
      final GetKeyValueCommand command =
         this.commandsFactory.buildGetKeyValueCommand(key, ctx.getFlags());
      final Object response = this.invoker.invoke(ctx, command);
      return response != null;
   }

   @Override
   public final boolean containsValue(Object value) {
      throw new UnsupportedOperationException("Not supported");
   }

   @Override
   public final V get(Object key) {
      return this.get(key, null, null);
   }

   @SuppressWarnings("unchecked")
   final V get(Object key, EnumSet<Flag> explicitFlags, ClassLoader explicitClassLoader) {
      this.assertKeyNotNull(key);
      final InvocationContext ctx =
         this.getInvocationContextForRead(null, explicitFlags, explicitClassLoader);
      final GetKeyValueCommand command =
         this.commandsFactory.buildGetKeyValueCommand(key, ctx.getFlags());
      return (V)this.invoker.invoke(ctx, command);
   }

   @Override
   public final V remove(Object key) {
      return this.remove(key, null, null);
   }

   @SuppressWarnings("unchecked")
   final V remove(Object key, EnumSet<Flag> explicitFlags, ClassLoader explicitClassLoader) {
      this.assertKeyNotNull(key);
      final InvocationContext ctx =
         this.getInvocationContextWithImplicitTransaction(explicitFlags, explicitClassLoader);
      final RemoveCommand command =
         this.commandsFactory.buildRemoveCommand(key, null, ctx.getFlags());
      return (V)this.executeCommandAndCommitIfNeeded(ctx, command);
   }

   @Override
   public final void clear() {
      this.clear(null, null);
   }

   final void clear(EnumSet<Flag> explicitFlags, ClassLoader explicitClassLoader) {
      final InvocationContext ctx =
         this.getInvocationContextWithImplicitTransaction(explicitFlags, explicitClassLoader);
      final ClearCommand command = this.commandsFactory.buildClearCommand(ctx.getFlags());
      this.executeCommandAndCommitIfNeeded(ctx, command);
   }

   @Override
   public Set<K> keySet() {
      return this.keySet(null, null);
   }

   @SuppressWarnings("unchecked")
   Set<K> keySet(EnumSet<Flag> explicitFlags, ClassLoader explicitClassLoader) {
      final InvocationContext ctx =
         this.getInvocationContextForRead(null, explicitFlags, explicitClassLoader);
      final KeySetCommand command = this.commandsFactory.buildKeySetCommand();
      return (Set<K>)this.invoker.invoke(ctx, command);
   }

   @Override
   public Collection<V> values() {
      return this.values(null, null);
   }

   @SuppressWarnings("unchecked")
   Collection<V> values(EnumSet<Flag> explicitFlags, ClassLoader explicitClassLoader) {
      final InvocationContext ctx =
         this.getInvocationContextForRead(null, explicitFlags, explicitClassLoader);
      final ValuesCommand command = this.commandsFactory.buildValuesCommand();
      return (Collection<V>)this.invoker.invoke(ctx, command);
   }

   @Override
   public Set<Map.Entry<K, V>> entrySet() {
      return this.entrySet(null, null);
   }

   @SuppressWarnings("unchecked")
   Set<Map.Entry<K, V>> entrySet(EnumSet<Flag> explicitFlags, ClassLoader explicitClassLoader) {
      final InvocationContext ctx =
         this.getInvocationContextForRead(null, explicitFlags, explicitClassLoader);
      final EntrySetCommand command = this.commandsFactory.buildEntrySetCommand();
      return (Set<Map.Entry<K, V>>)this.invoker.invoke(ctx, command);
   }

   @Override
   public final void putForExternalRead(K key, V value) {
      this.putForExternalRead(key, value, null, null);
   }

   final void putForExternalRead(K key, V value, EnumSet<Flag> explicitFlags,
      ClassLoader explicitClassLoader)
   {
      // Removed the "logic" to suspend the transaction around that put from the original code.
      // we use the cache as 100% transaction and it does not make sense for us to create 1
      // transaction per put and "auto commit" it.
      try {

         final EnumSet<Flag> flags =
            EnumSet.of(FAIL_SILENTLY, FORCE_ASYNCHRONOUS, ZERO_LOCK_ACQUISITION_TIMEOUT,
               PUT_FOR_EXTERNAL_READ);
         if (explicitFlags != null && !explicitFlags.isEmpty()) {
            flags.addAll(explicitFlags);
         }

         // if the entry exists then this should be a no-op.
         this.putIfAbsent(key, value, this.defaultLifespan, TimeUnit.MILLISECONDS,
            this.defaultMaxIdleTime, TimeUnit.MILLISECONDS, flags, explicitClassLoader);
      }
      catch (final Exception e) {
         if (CacheImpl.log.isDebugEnabled())
            CacheImpl.log.debug("Caught exception while doing putForExternalRead()", e);
      }
   }

   @Override
   public final void evict(K key) {
      this.evict(key, null, null);
   }

   final void evict(K key, EnumSet<Flag> explicitFlags, ClassLoader explicitClassLoader) {
      this.assertKeyNotNull(key);
      final InvocationContext ctx =
         this.createNonTxInvocationContext(explicitFlags, explicitClassLoader);
      final EvictCommand command = this.commandsFactory.buildEvictCommand(key);
      this.invoker.invoke(ctx, command);
   }

   private InvocationContext createNonTxInvocationContext(EnumSet<Flag> explicitFlags,
      ClassLoader explicitClassLoader)
   {
      final InvocationContext ctx = this.icc.createNonTxInvocationContext();
      return this.setInvocationContextFlagsAndClassLoader(ctx, explicitFlags, explicitClassLoader);
   }

   @Override
   public Configuration getConfiguration() {
      return this.config;
   }

   @Override
   public void addListener(Object listener) {
      this.notifier.addListener(listener);
   }

   @Override
   public void removeListener(Object listener) {
      this.notifier.removeListener(listener);
   }

   @Override
   public Set<Object> getListeners() {
      return this.notifier.getListeners();
   }

   private InvocationContext getInvocationContextForWrite(EnumSet<Flag> explicitFlags,
      ClassLoader explicitClassLoader)
   {
      final InvocationContext ctx = this.icc.createInvocationContext(true);
      return this.setInvocationContextFlagsAndClassLoader(ctx, explicitFlags, explicitClassLoader);
   }

   private InvocationContext getInvocationContextForRead(Transaction tx,
      EnumSet<Flag> explicitFlags, ClassLoader explicitClassLoader)
   {
      if (this.config.isTransactionalCache()) {
         final Transaction transaction = tx == null ? this.getOngoingTransaction() : tx;
         // if we are in the scope of a transaction than return a transactional context. This is
         // relevant e.g.
         // FORCE_WRITE_LOCK is used on read operations - in that case, the lock is held for the the
         // transaction's
         // lifespan (when in tx scope) vs. call lifespan (when not in tx scope).
         if (transaction != null)
            return this.getInvocationContext(transaction, explicitFlags, explicitClassLoader);
      }
      final InvocationContext result = this.icc.createInvocationContext(false);
      this.setInvocationContextFlagsAndClassLoader(result, explicitFlags, explicitClassLoader);
      return result;
   }

   /**
    * If this is a transactional cache and autoCommit is set to true then starts a transaction if
    * this is not a transactional call.
    */
   private InvocationContext getInvocationContextWithImplicitTransaction(
      EnumSet<Flag> explicitFlags, ClassLoader explicitClassLoader)
   {
      InvocationContext invocationContext;
      boolean txInjected = false;
      if (this.config.isTransactionalCache()) {
         Transaction transaction = this.getOngoingTransaction();
         if (transaction == null && this.config.isTransactionAutoCommit()) {
            try {
               this.transactionManager.begin();
               transaction = this.transactionManager.getTransaction();
               txInjected = true;
               if (CacheImpl.trace) CacheImpl.log.trace("Implicit transaction started!");
            }
            catch (final Exception e) {
               throw new CacheException("Could not start transaction", e);
            }
         }
         invocationContext =
            this.getInvocationContext(transaction, explicitFlags, explicitClassLoader);
      }
      else {
         invocationContext = this.getInvocationContextForWrite(explicitFlags, explicitClassLoader);
      }
      if (txInjected) {
         ((TxInvocationContext)invocationContext).setImplicitTransaction(true);
         if (CacheImpl.trace) CacheImpl.log.tracef("Marked tx as implicit.");
      }
      return invocationContext;
   }

   private InvocationContext getInvocationContext(Transaction tx, EnumSet<Flag> explicitFlags,
      ClassLoader explicitClassLoader)
   {
      final InvocationContext ctx = this.icc.createInvocationContext(tx);
      return this.setInvocationContextFlagsAndClassLoader(ctx, explicitFlags, explicitClassLoader);
   }

   private InvocationContext setInvocationContextFlagsAndClassLoader(InvocationContext ctx,
      EnumSet<Flag> explicitFlags, ClassLoader explicitClassLoader)
   {
      if (explicitFlags != null) ctx.setFlags(explicitFlags);
      ctx.setClassLoader(explicitClassLoader != null ?
         explicitClassLoader : this.getClassLoader());
      return ctx;
   }

   @Override
   public boolean lock(K... keys) {
      this.assertKeyNotNull(keys);
      return this.lock(Arrays.asList(keys), null, null);
   }

   @Override
   public boolean lock(Collection<? extends K> keys) {
      return this.lock(keys, null, null);
   }

   boolean lock(Collection<? extends K> keys, EnumSet<Flag> explicitFlags,
      ClassLoader explicitClassLoader)
   {
      if (keys == null || keys.isEmpty()) {
         throw new IllegalArgumentException("Cannot lock empty list of keys");
      }
      final InvocationContext ctx =
         this.getInvocationContextForWrite(explicitFlags, explicitClassLoader);
      final LockControlCommand command =
         this.commandsFactory.buildLockControlCommand(keys, ctx.getFlags());
      return (Boolean)this.invoker.invoke(ctx, command);
   }

   @Override
   public void applyDelta(K deltaAwareValueKey, Delta delta, Object... locksToAcquire) {
      if (locksToAcquire == null || locksToAcquire.length == 0) {
         throw new IllegalArgumentException("Cannot lock empty list of keys");
      }
      final InvocationContext ctx = this.getInvocationContextForWrite(null, null);
      final ApplyDeltaCommand command =
         this.commandsFactory.buildApplyDeltaCommand(deltaAwareValueKey, delta,
            Arrays.asList(locksToAcquire));
      this.invoker.invoke(ctx, command);
   }

   @Override
   @ManagedOperation(description = "Starts the cache.")
   @Operation(displayName = "Starts cache.")
   public void start() {
      this.componentRegistry.start();
      this.defaultLifespan = this.config.getExpirationLifespan();
      this.defaultMaxIdleTime = this.config.getExpirationMaxIdle();
      if (CacheImpl.log.isDebugEnabled())
         CacheImpl.log.debugf("Started cache %s on %s", this.getName(), this.getCacheManager()
            .getAddress());
   }

   @Override
   @ManagedOperation(description = "Stops the cache.")
   @Operation(displayName = "Stops cache.")
   public void stop() {
      this.stop(null, null);
   }

   void stop(EnumSet<Flag> explicitFlags, ClassLoader explicitClassLoader) {
      if (CacheImpl.log.isDebugEnabled())
         CacheImpl.log.debugf("Stopping cache %s on %s", this.getName(), this.getCacheManager()
            .getAddress());

      // Create invocation context to pass flags
      this.createNonTxInvocationContext(explicitFlags, explicitClassLoader);
      this.componentRegistry.stop();
   }

   @Override
   public List<CommandInterceptor> getInterceptorChain() {
      return this.invoker.asList();
   }

   @Override
   public void addInterceptor(CommandInterceptor i, int position) {
      this.invoker.addInterceptor(i, position);
   }

   @Override
   public boolean addInterceptorAfter(CommandInterceptor i,
      Class<? extends CommandInterceptor> afterInterceptor)
   {
      return this.invoker.addInterceptorAfter(i, afterInterceptor);
   }

   @Override
   public boolean addInterceptorBefore(CommandInterceptor i,
      Class<? extends CommandInterceptor> beforeInterceptor)
   {
      return this.invoker.addInterceptorBefore(i, beforeInterceptor);
   }

   @Override
   public void removeInterceptor(int position) {
      this.invoker.removeInterceptor(position);
   }

   @Override
   public void removeInterceptor(Class<? extends CommandInterceptor> interceptorType) {
      this.invoker.removeInterceptor(interceptorType);
   }

   @Override
   public EvictionManager getEvictionManager() {
      return this.evictionManager;
   }

   @Override
   public ComponentRegistry getComponentRegistry() {
      return this.componentRegistry;
   }

   @Override
   public DistributionManager getDistributionManager() {
      return this.distributionManager;
   }

   @Override
   public ComponentStatus getStatus() {
      return this.componentRegistry.getStatus();
   }

   /**
    * Returns String representation of ComponentStatus enumeration in order to avoid class not found
    * exceptions in JMX tools that don't have access to infinispan classes.
    */
   @ManagedAttribute(description = "Returns the cache status")
   @Metric(displayName = "Cache status", dataType = DataType.TRAIT,
      displayType = DisplayType.SUMMARY)
   public String getCacheStatus() {
      return this.getStatus().toString();
   }

   @Override
   public boolean startBatch() {
      if (!this.config.isInvocationBatchingEnabled()) {
         throw new ConfigurationException(
            "Invocation batching not enabled in current configuration!  Please use the <invocationBatching /> element.");
      }
      return this.batchContainer.startBatch();
   }

   @Override
   public void endBatch(boolean successful) {
      if (!this.config.isInvocationBatchingEnabled()) {
         throw new ConfigurationException(
            "Invocation batching not enabled in current configuration!  Please use the <invocationBatching /> element.");
      }
      this.batchContainer.endBatch(successful);
   }

   @Override
   public String getName() {
      return this.name;
   }

   /**
    * Returns the cache name. If this is the default cache, it returns a more friendly name.
    */
   @ManagedAttribute(description = "Returns the cache name")
   @Metric(displayName = "Cache name", dataType = DataType.TRAIT, displayType = DisplayType.SUMMARY)
   public String getCacheName() {
      final String name =
         this.getName().equals(BasicCacheContainer.DEFAULT_CACHE_NAME) ? "Default Cache" : this
            .getName();
      return name + "(" + this.getConfiguration().getCacheModeString().toLowerCase() + ")";
   }

   /**
    * Returns the cache configuration as XML string.
    */
   @ManagedAttribute(description = "Returns the cache configuration as XML string")
   @Metric(displayName = "Cache configuration (XML)", dataType = DataType.TRAIT,
      displayType = DisplayType.SUMMARY)
   public String getConfigurationAsXmlString() {
      return this.getConfiguration().toXmlString();
   }

   @Override
   public String getVersion() {
      return Version.VERSION;
   }

   @Override
   public String toString() {
      return "Cache '" +
         this.name +
         "'@" +
         (this.config.getCacheMode().isClustered() ? this.getCacheManager().getAddress() : Util
            .hexIdHashCode(this));
   }

   @Override
   public BatchContainer getBatchContainer() {
      return this.batchContainer;
   }

   @Override
   public InvocationContextContainer getInvocationContextContainer() {
      return this.icc;
   }

   @Override
   public DataContainer getDataContainer() {
      return this.dataContainer;
   }

   @Override
   public TransactionManager getTransactionManager() {
      return this.transactionManager;
   }

   @Override
   public LockManager getLockManager() {
      return this.lockManager;
   }

   @Override
   public EmbeddedCacheManager getCacheManager() {
      return this.cacheManager;
   }

   @Override
   public Stats getStats() {
      return new StatsImpl(this.invoker);
   }

   @Override
   public XAResource getXAResource() {
      return new TransactionXaAdapter(this.txTable, this.config, this.recoveryManager,
         this.txCoordinator, this.commandsFactory, this.rpcManager, null, this.config);
   }

   @Override
   public final V put(K key, V value, long lifespan, TimeUnit lifespanUnit, long maxIdleTime,
      TimeUnit idleTimeUnit)
   {
      return this.put(key, value, lifespan, lifespanUnit, maxIdleTime, idleTimeUnit, null, null);
   }

   @SuppressWarnings("unchecked")
   final V put(K key, V value, long lifespan, TimeUnit lifespanUnit, long maxIdleTime,
      TimeUnit idleTimeUnit, EnumSet<Flag> explicitFlags, ClassLoader explicitClassLoader)
   {
      this.assertKeyNotNull(key);
      final InvocationContext ctx =
         this.getInvocationContextWithImplicitTransaction(explicitFlags, explicitClassLoader);
      final PutKeyValueCommand command =
         this.commandsFactory.buildPutKeyValueCommand(key, value, lifespanUnit.toMillis(lifespan),
            idleTimeUnit.toMillis(maxIdleTime), ctx.getFlags());
      return (V)this.executeCommandAndCommitIfNeeded(ctx, command);
   }

   @Override
   public final V putIfAbsent(K key, V value, long lifespan, TimeUnit lifespanUnit,
      long maxIdleTime, TimeUnit idleTimeUnit)
   {
      return this.putIfAbsent(key, value, lifespan, lifespanUnit, maxIdleTime, idleTimeUnit, null,
         null);
   }

   @SuppressWarnings("unchecked")
   final V putIfAbsent(K key, V value, long lifespan, TimeUnit lifespanUnit, long maxIdleTime,
      TimeUnit idleTimeUnit, EnumSet<Flag> explicitFlags, ClassLoader explicitClassLoader)
   {
      this.assertKeyNotNull(key);
      final InvocationContext ctx =
         this.getInvocationContextWithImplicitTransaction(explicitFlags, explicitClassLoader);
      final PutKeyValueCommand command =
         this.commandsFactory.buildPutKeyValueCommand(key, value, lifespanUnit.toMillis(lifespan),
            idleTimeUnit.toMillis(maxIdleTime), ctx.getFlags());
      command.setPutIfAbsent(true);
      return (V)this.executeCommandAndCommitIfNeeded(ctx, command);
   }

   @Override
   public final void putAll(Map<? extends K, ? extends V> map, long lifespan, TimeUnit lifespanUnit,
      long maxIdleTime, TimeUnit idleTimeUnit)
   {
      this.putAll(map, lifespan, lifespanUnit, maxIdleTime, idleTimeUnit, null, null);
   }

   final void putAll(Map<? extends K, ? extends V> map, long lifespan, TimeUnit lifespanUnit,
      long maxIdleTime, TimeUnit idleTimeUnit, EnumSet<Flag> explicitFlags,
      ClassLoader explicitClassLoader)
   {
      this.assertKeysNotNull(map);
      final InvocationContext ctx =
         this.getInvocationContextWithImplicitTransaction(explicitFlags, explicitClassLoader);
      final PutMapCommand command =
         this.commandsFactory.buildPutMapCommand(map, lifespanUnit.toMillis(lifespan),
            idleTimeUnit.toMillis(maxIdleTime), ctx.getFlags());
      this.executeCommandAndCommitIfNeeded(ctx, command);
   }

   @Override
   public final V replace(K key, V value, long lifespan, TimeUnit lifespanUnit, long maxIdleTime,
      TimeUnit idleTimeUnit)
   {
      return this.replace(key, value, lifespan, lifespanUnit, maxIdleTime, idleTimeUnit, null, null);
   }

   @SuppressWarnings("unchecked")
   final V replace(K key, V value, long lifespan, TimeUnit lifespanUnit, long maxIdleTime,
      TimeUnit idleTimeUnit, EnumSet<Flag> explicitFlags, ClassLoader explicitClassLoader)
   {
      this.assertKeyNotNull(key);
      final InvocationContext ctx =
         this.getInvocationContextWithImplicitTransaction(explicitFlags, explicitClassLoader);
      final ReplaceCommand command =
         this.commandsFactory.buildReplaceCommand(key, null, value, lifespanUnit.toMillis(lifespan),
            idleTimeUnit.toMillis(maxIdleTime), ctx.getFlags());
      return (V)this.executeCommandAndCommitIfNeeded(ctx, command);

   }

   @Override
   public final boolean replace(K key, V oldValue, V value, long lifespan, TimeUnit lifespanUnit,
      long maxIdleTime, TimeUnit idleTimeUnit)
   {
      return this.replace(key, oldValue, value, lifespan, lifespanUnit, maxIdleTime, idleTimeUnit,
         null, null);
   }

   final boolean replace(K key, V oldValue, V value, long lifespan, TimeUnit lifespanUnit,
      long maxIdleTime, TimeUnit idleTimeUnit, EnumSet<Flag> explicitFlags,
      ClassLoader explicitClassLoader)
   {
      this.assertKeyNotNull(key);
      final InvocationContext ctx =
         this.getInvocationContextWithImplicitTransaction(explicitFlags, explicitClassLoader);
      final ReplaceCommand command =
         this.commandsFactory.buildReplaceCommand(key, oldValue, value,
            lifespanUnit.toMillis(lifespan), idleTimeUnit.toMillis(maxIdleTime), ctx.getFlags());
      return (Boolean)this.executeCommandAndCommitIfNeeded(ctx, command);
   }

   /**
    * Wraps a return value as a future, if needed. Typically, if the stack, operation and
    * configuration support handling of futures, this retval is already a future in which case this
    * method does nothing except cast to future.
    * <p/>
    * Otherwise, a future wrapper is created, which has already completed and simply returns the
    * retval. This is used for API consistency.
    * 
    * @param retval return value to wrap
    * @param <X> contents of the future
    * @return a future
    */
   @SuppressWarnings("unchecked")
   private <X> NotifyingFuture<X> wrapInFuture(final Object retval) {
      if (retval instanceof NotifyingFuture) {
         return (NotifyingFuture<X>)retval;
      }
      else {
         return new AbstractInProcessNotifyingFuture<X>() {
            @Override
            @SuppressWarnings("unchecked")
            public X get() throws InterruptedException, ExecutionException {
               return (X)retval;
            }
         };
      }
   }

   @Override
   public final NotifyingFuture<V> putAsync(K key, V value, long lifespan, TimeUnit lifespanUnit,
      long maxIdle, TimeUnit maxIdleUnit)
   {
      return this.putAsync(key, value, lifespan, lifespanUnit, maxIdle, maxIdleUnit, null, null);
   }

   final NotifyingFuture<V> putAsync(K key, V value, long lifespan, TimeUnit lifespanUnit,
      long maxIdle, TimeUnit maxIdleUnit, EnumSet<Flag> explicitFlags,
      ClassLoader explicitClassLoader)
   {
      this.assertKeyNotNull(key);
      final InvocationContext ctx =
         this.getInvocationContextWithImplicitTransaction(explicitFlags, explicitClassLoader);
      ctx.setUseFutureReturnType(true);
      final PutKeyValueCommand command =
         this.commandsFactory.buildPutKeyValueCommand(key, value, lifespanUnit.toMillis(lifespan),
            maxIdleUnit.toMillis(maxIdle), ctx.getFlags());
      return this.wrapInFuture(this.executeCommandAndCommitIfNeeded(ctx, command));
   }

   @Override
   public final NotifyingFuture<Void> putAllAsync(Map<? extends K, ? extends V> data, long lifespan,
      TimeUnit lifespanUnit, long maxIdle, TimeUnit maxIdleUnit)
   {
      return this.putAllAsync(data, lifespan, lifespanUnit, maxIdle, maxIdleUnit, null, null);
   }

   final NotifyingFuture<Void> putAllAsync(Map<? extends K, ? extends V> data, long lifespan,
      TimeUnit lifespanUnit, long maxIdle, TimeUnit maxIdleUnit, EnumSet<Flag> explicitFlags,
      ClassLoader explicitClassLoader)
   {
      this.assertKeysNotNull(data);
      final InvocationContext ctx =
         this.getInvocationContextWithImplicitTransaction(explicitFlags, explicitClassLoader);
      ctx.setUseFutureReturnType(true);
      final PutMapCommand command =
         this.commandsFactory.buildPutMapCommand(data, lifespanUnit.toMillis(lifespan),
            maxIdleUnit.toMillis(maxIdle), ctx.getFlags());
      return this.wrapInFuture(this.executeCommandAndCommitIfNeeded(ctx, command));
   }

   @Override
   public final NotifyingFuture<Void> clearAsync() {
      return this.clearAsync(null, null);
   }

   final NotifyingFuture<Void> clearAsync(EnumSet<Flag> explicitFlags,
      ClassLoader explicitClassLoader)
   {
      final InvocationContext ctx =
         this.getInvocationContextWithImplicitTransaction(explicitFlags, explicitClassLoader);
      ctx.setUseFutureReturnType(true);
      final ClearCommand command = this.commandsFactory.buildClearCommand(ctx.getFlags());
      return this.wrapInFuture(this.executeCommandAndCommitIfNeeded(ctx, command));
   }

   @Override
   public final NotifyingFuture<V> putIfAbsentAsync(K key, V value, long lifespan,
      TimeUnit lifespanUnit, long maxIdle, TimeUnit maxIdleUnit)
   {
      return this.putIfAbsentAsync(key, value, lifespan, lifespanUnit, maxIdle, maxIdleUnit, null,
         null);
   }

   final NotifyingFuture<V> putIfAbsentAsync(K key, V value, long lifespan, TimeUnit lifespanUnit,
      long maxIdle, TimeUnit maxIdleUnit, EnumSet<Flag> explicitFlags,
      ClassLoader explicitClassLoader)
   {
      this.assertKeyNotNull(key);
      final InvocationContext ctx =
         this.getInvocationContextWithImplicitTransaction(explicitFlags, explicitClassLoader);
      ctx.setUseFutureReturnType(true);
      final PutKeyValueCommand command =
         this.commandsFactory.buildPutKeyValueCommand(key, value, lifespanUnit.toMillis(lifespan),
            maxIdleUnit.toMillis(maxIdle), ctx.getFlags());
      command.setPutIfAbsent(true);
      return this.wrapInFuture(this.executeCommandAndCommitIfNeeded(ctx, command));
   }

   @Override
   public final NotifyingFuture<V> removeAsync(Object key) {
      return this.removeAsync(key, null, null);
   }

   final NotifyingFuture<V> removeAsync(Object key, EnumSet<Flag> explicitFlags,
      ClassLoader explicitClassLoader)
   {
      this.assertKeyNotNull(key);
      final InvocationContext ctx =
         this.getInvocationContextWithImplicitTransaction(explicitFlags, explicitClassLoader);
      ctx.setUseFutureReturnType(true);
      final RemoveCommand command =
         this.commandsFactory.buildRemoveCommand(key, null, ctx.getFlags());
      return this.wrapInFuture(this.executeCommandAndCommitIfNeeded(ctx, command));
   }

   @Override
   public final NotifyingFuture<Boolean> removeAsync(Object key, Object value) {
      return this.removeAsync(key, value, null, null);
   }

   final NotifyingFuture<Boolean> removeAsync(Object key, Object value, EnumSet<Flag> explicitFlags,
      ClassLoader explicitClassLoader)
   {
      this.assertKeyNotNull(key);
      final InvocationContext ctx =
         this.getInvocationContextWithImplicitTransaction(explicitFlags, explicitClassLoader);
      ctx.setUseFutureReturnType(true);
      final RemoveCommand command =
         this.commandsFactory.buildRemoveCommand(key, value, ctx.getFlags());
      return this.wrapInFuture(this.executeCommandAndCommitIfNeeded(ctx, command));
   }

   @Override
   public final NotifyingFuture<V> replaceAsync(K key, V value, long lifespan,
      TimeUnit lifespanUnit, long maxIdle, TimeUnit maxIdleUnit)
   {
      return this.replaceAsync(key, value, lifespan, lifespanUnit, maxIdle, maxIdleUnit, null, null);
   }

   final NotifyingFuture<V> replaceAsync(K key, V value, long lifespan, TimeUnit lifespanUnit,
      long maxIdle, TimeUnit maxIdleUnit, EnumSet<Flag> explicitFlags,
      ClassLoader explicitClassLoader)
   {
      this.assertKeyNotNull(key);
      final InvocationContext ctx =
         this.getInvocationContextWithImplicitTransaction(explicitFlags, explicitClassLoader);
      ctx.setUseFutureReturnType(true);
      final ReplaceCommand command =
         this.commandsFactory.buildReplaceCommand(key, null, value, lifespanUnit.toMillis(lifespan),
            maxIdleUnit.toMillis(maxIdle), ctx.getFlags());
      return this.wrapInFuture(this.executeCommandAndCommitIfNeeded(ctx, command));
   }

   @Override
   public final NotifyingFuture<Boolean> replaceAsync(K key, V oldValue, V newValue, long lifespan,
      TimeUnit lifespanUnit, long maxIdle, TimeUnit maxIdleUnit)
   {
      return this.replaceAsync(key, oldValue, newValue, lifespan, lifespanUnit, maxIdle,
         maxIdleUnit, null, null);
   }

   final NotifyingFuture<Boolean> replaceAsync(K key, V oldValue, V newValue, long lifespan,
      TimeUnit lifespanUnit, long maxIdle, TimeUnit maxIdleUnit, EnumSet<Flag> explicitFlags,
      ClassLoader explicitClassLoader)
   {
      this.assertKeyNotNull(key);
      final InvocationContext ctx =
         this.getInvocationContextWithImplicitTransaction(explicitFlags, explicitClassLoader);
      ctx.setUseFutureReturnType(true);
      final ReplaceCommand command =
         this.commandsFactory.buildReplaceCommand(key, oldValue, newValue,
            lifespanUnit.toMillis(lifespan), maxIdleUnit.toMillis(maxIdle), ctx.getFlags());
      return this.wrapInFuture(this.executeCommandAndCommitIfNeeded(ctx, command));
   }

   @Override
   public NotifyingFuture<V> getAsync(K key) {
      return this.getAsync(key, null, null);
   }

   @SuppressWarnings("unchecked")
   NotifyingFuture<V> getAsync(final K key, final EnumSet<Flag> explicitFlags,
      final ClassLoader explicitClassLoader)
   {
      final Transaction tx = this.getOngoingTransaction();
      final NotifyingNotifiableFuture f = new DeferredReturnFuture();

      // Optimization to not start a new thread only when the operation is cheap:
      if (this.asyncSkipsThread(explicitFlags, key)) {
         return this.wrapInFuture(this.get(key, explicitFlags, explicitClassLoader));
      }
      else {
         // Make sure the flags are cleared
         final EnumSet<Flag> appliedFlags;
         if (explicitFlags == null) {
            appliedFlags = null;
         }
         else {
            appliedFlags = explicitFlags.clone();
            explicitFlags.clear();
         }
         final Callable<V> c = new Callable<V>() {
            @Override
            public V call() throws Exception {
               CacheImpl.this.assertKeyNotNull(key);
               final InvocationContext ctx =
                  CacheImpl.this.getInvocationContextForRead(tx, appliedFlags, explicitClassLoader);
               final GetKeyValueCommand command =
                  CacheImpl.this.commandsFactory.buildGetKeyValueCommand(key, appliedFlags);
               final Object ret = CacheImpl.this.invoker.invoke(ctx, command);
               f.notifyDone();
               return (V)ret;
            }
         };
         f.setNetworkFuture(this.asyncExecutor.submit(c));
         return f;
      }
   }

   /**
    * Encodes the cases for an asyncGet operation in which it makes sense to actually perform the
    * operation in sync.
    * 
    * @param flags
    * @param key
    * @return true if we skip the thread (performing it in sync)
    */
   private boolean asyncSkipsThread(EnumSet<Flag> flags, K key) {
      final boolean isSkipLoader = this.isSkipLoader(flags);
      if (!isSkipLoader) {
         // if we can't skip the cacheloader, we really want a thread for async.
         return false;
      }
      final CacheMode cacheMode = this.config.getCacheMode();
      if (!cacheMode.isDistributed()) {
         // in these cluster modes we won't RPC for a get, so no need to fork a thread.
         return true;
      }
      else if (flags != null &&
         (flags.contains(Flag.SKIP_REMOTE_LOOKUP) || flags.contains(Flag.CACHE_MODE_LOCAL))) {
         // with these flags we won't RPC either
         return true;
      }
      // finally, we will skip the thread if the key maps to the local node
      return this.distributionManager.getLocality(key).isLocal();
   }

   private boolean isSkipLoader(EnumSet<Flag> flags) {
      final boolean hasCacheLoaderConfig =
         this.config.getCacheLoaderManagerConfig().getFirstCacheLoaderConfig() != null;
      return !hasCacheLoaderConfig
         ||
         (hasCacheLoaderConfig && flags != null && (flags.contains(Flag.SKIP_CACHE_LOAD) || flags
            .contains(Flag.SKIP_CACHE_STORE)));
   }

   @Override
   public AdvancedCache<K, V> getAdvancedCache() {
      return this;
   }

   @Override
   public void compact() {
      for (final InternalCacheEntry e : this.dataContainer) {
         if (e.getKey() instanceof MarshalledValue) {
            ((MarshalledValue)e.getKey()).compact(true, true);
         }
         if (e.getValue() instanceof MarshalledValue) {
            ((MarshalledValue)e.getValue()).compact(true, true);
         }
      }
   }

   @Override
   public RpcManager getRpcManager() {
      return this.rpcManager;
   }

   @Override
   public AdvancedCache<K, V> withFlags(Flag... flags) {
      if (flags == null || flags.length == 0) return this;
      else return new DecoratedCache<K, V>(this, flags);
   }

   private Transaction getOngoingTransaction() {
      try {
         Transaction transaction = null;
         if (this.transactionManager != null) {
            transaction = this.transactionManager.getTransaction();
            if (transaction == null && this.config.isInvocationBatchingEnabled()) {
               transaction = this.batchContainer.getBatchTransaction();
            }
         }
         return transaction;
      }
      catch (final SystemException e) {
         throw new CacheException("Unable to get transaction", e);
      }
   }

   private Object executeCommandAndCommitIfNeeded(InvocationContext ctx, VisitableCommand command) {
      final boolean txInjected =
         ctx.isInTxScope() && ((TxInvocationContext)ctx).isImplicitTransaction();
      Object result;
      try {
         result = this.invoker.invoke(ctx, command);
      }
      catch (final RuntimeException e) {
         if (txInjected) this.tryRollback();
         throw e;
      }

      if (txInjected) {
         if (CacheImpl.trace)
            CacheImpl.log.tracef("Committing transaction as it was implicit: %s",
               this.getOngoingTransaction());
         try {
            this.transactionManager.commit();
         }
         catch (final Throwable e) {
            CacheImpl.log.couldNotCompleteInjectedTransaction(e);
            this.tryRollback();
            throw new CacheException("Could not commit implicit transaction", e);
         }
      }

      return result;
   }

   private void tryRollback() {
      try {
         if (this.transactionManager != null) this.transactionManager.rollback();
      }
      catch (final Throwable t) {
         if (CacheImpl.trace) CacheImpl.log.trace("Could not rollback", t);// best effort
      }
   }

   @Override
   public ClassLoader getClassLoader() {
      return this.config.getClassLoader();
   }

   @Override
   public AdvancedCache<K, V> with(ClassLoader classLoader) {
      return new DecoratedCache<K, V>(this, classLoader);
   }

   @Override
   protected void set(K key, V value) {
      this.withFlags(Flag.SKIP_REMOTE_LOOKUP, Flag.SKIP_CACHE_LOAD)
         .put(key, value, this.defaultLifespan, MILLISECONDS, this.defaultMaxIdleTime, MILLISECONDS);
   }
}
